import dataclasses
from functools import cached_property
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Callable,
    Dict,
    List,
    Literal,
    Mapping,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
    cast,
    get_args,
    overload,
)

from django.db import models
from django.db.models import QuerySet
from django.db.models.fields.related_descriptors import (
    ForwardManyToOneDescriptor,
    ReverseManyToOneDescriptor,
    ReverseOneToOneDescriptor,
)
from django.db.models.query_utils import DeferredAttribute
import strawberry
from strawberry import UNSET
from strawberry.annotation import StrawberryAnnotation
from strawberry.arguments import StrawberryArgument
from strawberry.lazy_type import LazyType
from strawberry.permission import BasePermission
from strawberry.type import StrawberryContainer, StrawberryType
from strawberry.types.fields.resolver import StrawberryResolver
from strawberry.types.info import Info
from strawberry.types.types import TypeDefinition
from strawberry.union import StrawberryUnion
from strawberry_django.arguments import argument
from strawberry_django.fields.field import (
    StrawberryDjangoField as _StrawberryDjangoField,
)
from strawberry_django.utils import get_django_model, unwrap_type

from . import relay
from .descriptors import ModelProperty
from .optimizer import OptimizerStore, PrefetchType
from .permissions import filter_with_perms
from .utils import resolvers
from .utils.typing import TypeOrSequence

if TYPE_CHECKING:
    pass

__all__ = [
    "StrawberryDjangoField",
    "StrawberryDjangoNodeField",
    "StrawberryDjangoConnectionField",
    "field",
    "node",
    "connection",
]

_T = TypeVar("_T")
_M = TypeVar("_M", bound=models.Model)


class StrawberryDjangoField(_StrawberryDjangoField):
    """A strawberry field for django attributes.

    Do not instantiate this directly. Instead, use `@field` decorator.

    """

    store: OptimizerStore

    def __init__(self, *args, **kwargs):
        self.disable_optimization = kwargs.pop("disable_optimization", False)
        self.store = OptimizerStore.with_hints(
            only=kwargs.pop("only", None),
            select_related=kwargs.pop("select_related", None),
            prefetch_related=kwargs.pop("prefetch_related", None),
        )
        super().__init__(*args, **kwargs)

    @cached_property
    def is_basic_field(self):
        return False

    @property
    def arguments(self) -> List[StrawberryArgument]:
        if isinstance(self, relay.NodeField):
            return []

        args = super().arguments
        is_node = isinstance(unwrap_type(self.type), relay.Node)
        return [
            (
                (
                    argument("ids", List[relay.GlobalID], is_optional=self.is_optional)
                    if self.is_list
                    else argument("id", relay.GlobalID, is_optional=self.is_optional)
                )
                if is_node
                and arg.python_name == "pk"
                and arg.type_annotation.annotation == strawberry.ID
                else arg
            )
            for arg in args
        ]

    @property
    def type(self) -> Union[StrawberryType, type]:  # noqa:A003
        return super().type

    @type.setter
    def type(self, type_: Any) -> None:  # noqa:A003
        if type_ is not None and self.origin_django_type is None:
            resolved = type_
            if isinstance(resolved, StrawberryAnnotation):
                resolved = type_.resolve()
            while isinstance(resolved, StrawberryContainer):
                resolved = resolved.of_type

            dj_type = getattr(resolved, "_django_type", None)
            if dj_type is None:
                contained = get_args(resolved)
                if contained:
                    dj_type = getattr(contained[0], "_django_type", None)

            if dj_type is not None:
                self.origin_django_type = dj_type
                if self.filters is UNSET or self.filters is None:
                    self.filters = dj_type.filters
                if self.order is UNSET or self.order is None:
                    self.order = dj_type.order
                if self.pagination is UNSET or self.pagination is None:
                    self.pagination = dj_type.pagination

        super(StrawberryDjangoField, self.__class__).type.fset(self, type_)  # type:ignore

    @cached_property
    def model(self) -> Type[models.Model]:
        type_ = unwrap_type(self.type)
        model = get_django_model(type_)
        if model:
            return model

        tdef = cast(Optional[TypeDefinition], getattr(type_, "_type_definition", None))
        if tdef and tdef.concrete_of and issubclass(tdef.concrete_of.origin, relay.Connection):
            n_type = tdef.type_var_map[relay.NodeType]  # type:ignore
            if isinstance(n_type, LazyType):
                n_type = n_type.resolve_type()

            return cast(Type[models.Model], get_django_model(n_type))

        origin = self.origin_django_type or getattr(self.origin, "_django_type", None)
        model = origin and origin.model

        if isinstance(self.type, StrawberryUnion):
            mlist = []
            for t in self.type.types:
                dj_type = getattr(t, "_django_type", None)
                if dj_type:
                    mlist.append(dj_type.model)
                else:
                    model = getattr(t, "model", None)
                    if model:
                        mlist.append(model)
            assert len(mlist) == 1
            model = mlist[0]

        return cast(Type[models.Model], model)

    @cached_property
    def safe_resolver(self):
        resolver = self.base_resolver
        assert resolver
        if not resolver.is_async:
            resolver = resolvers.async_safe(resolver)
        return resolver

    def get_result(
        self,
        source: Optional[models.Model],
        info: Info,
        args: List[Any],
        kwargs: Dict[str, Any],
    ) -> Union[Awaitable[Any], Any]:
        if self.base_resolver is not None:
            result = self.resolver(source, info, args, kwargs)
        elif source is None:
            result = self.model._default_manager.all()
        else:
            # Small optimization to async resolvers avoid having to call it in an sync_to_async
            # context if the value is already cached, since it will not hit the db anymore
            attname = self.django_name or self.python_name
            attr = getattr(source.__class__, attname, None)
            try:
                if isinstance(attr, ModelProperty):
                    result = source.__dict__[attr.name]
                elif isinstance(attr, DeferredAttribute):
                    # If the value is cached, retrieve it with getattr because
                    # some fields wrap values at that time (e.g. FileField).
                    # If this next like fails, it will raise KeyError and get
                    # us out of the loop before we can do getattr
                    source.__dict__[attr.field.attname]
                    result = getattr(source, attr.field.attname)
                elif isinstance(attr, ForwardManyToOneDescriptor):
                    # This will raise KeyError if it is not cached
                    result = attr.field.get_cached_value(source)  # type:ignore
                elif isinstance(attr, ReverseOneToOneDescriptor):
                    # This will raise KeyError if it is not cached
                    result = attr.related.get_cached_value(source)
                elif isinstance(attr, ReverseManyToOneDescriptor):
                    # This returns a queryset, it is async safe
                    result = getattr(source, attname)
                else:
                    raise KeyError
            except KeyError:
                result = resolvers.getattr_async_safe(source, attname)

        if self.is_list:
            qs_resolver = lambda qs: self.get_queryset_as_list(qs, info, kwargs)
        else:
            qs_resolver = lambda qs: self.get_queryset_one(qs, info, kwargs)

        return resolvers.resolve_result(result, info=info, qs_resolver=qs_resolver)

    def resolver(
        self,
        source: Any,
        info: Info,
        args: List[Any],
        kwargs: Dict[str, Any],
    ) -> Any:
        return self.safe_resolver(*args, **kwargs)

    @resolvers.async_safe
    def get_queryset_as_list(
        self,
        qs: QuerySet[_M],
        info: Info,
        kwargs: Dict[str, Any],
        *,
        skip_fetch: bool = False,
    ) -> QuerySet[_M]:
        # Remove info from kwargs since we will pass it positionaly to get_queryset
        if "info" in kwargs:
            del kwargs["info"]

        if not self.base_resolver:
            nodes: Optional[List[relay.GlobalID]] = kwargs.get("ids")
            if isinstance(nodes, list):
                if nodes:
                    assert {n.resolve_type(info) for n in nodes} == {unwrap_type(self.type)}
                qs = qs.filter(pk__in=[n.node_id for n in nodes])

        qs = self.get_queryset(filter_with_perms(qs, info), info, **kwargs)
        if not skip_fetch and not isinstance(self, relay.ConnectionField):
            # This is what QuerySet does internally to fetch results.
            # After this, iterating over the queryset should be async safe
            qs._fetch_all()  # type:ignore
        return qs

    @resolvers.async_safe
    def get_queryset_one(
        self,
        qs: QuerySet[_M],
        info: Info,
        kwargs: Dict[str, Any],
    ) -> Optional[_M]:
        # Remove info from kwargs since we will pass it positionaly to get_queryset
        if "info" in kwargs:
            del kwargs["info"]

        try:
            qs = self.get_queryset(qs, info, **kwargs)
            if not self.base_resolver:
                node = kwargs.get("id")
                if isinstance(node, relay.GlobalID):
                    assert node.resolve_type(info) == unwrap_type(self.type)
                    qs = qs.filter(pk=node.node_id)
            return qs.get()
        except self.model.DoesNotExist:
            if not self.is_optional:
                raise

        return None


class StrawberryDjangoNodeField(relay.NodeField, StrawberryDjangoField):
    ...


class StrawberryDjangoConnectionField(relay.ConnectionField, StrawberryDjangoField):
    def __init__(self, *args, **kwargs):
        self._force_is_list_as_true = False
        super().__init__(*args, **kwargs)

    @property
    def arguments(self) -> List[StrawberryArgument]:
        # FIXME: The order/filters/etc arguments will only be added if there's no base_resolver
        # defined, but for Connections the resolver will resolve the iterable and not the field
        # return value itself. How to handle this in a better way?
        base_resolver = self._base_resolver
        self._base_resolver = None
        self._force_is_list_as_true = True
        args = super().arguments
        if base_resolver is not None:
            args += base_resolver.arguments
        self._base_resolver = base_resolver
        self._force_is_list_as_true = False
        return args

    def is_list(self):
        if self._force_is_list_as_true:
            return True
        return super().is_list

    def resolve_nodes(
        self,
        source: Any,
        info: Info,
        args: List[Any],
        kwargs: Dict[str, Any],
        *,
        nodes: Optional[QuerySet[Any]] = None,
    ):
        if nodes is None:
            nodes = self.model._default_manager.all()

            if self.origin_django_type and hasattr(self.origin_django_type.origin, "get_queryset"):
                nodes = cast(
                    QuerySet[Any],
                    self.origin_django_type.origin.get_queryset(nodes, info),
                )

        return self.get_queryset_as_list(nodes, info, kwargs, skip_fetch=True)


@overload
def field(
    *,
    resolver: Callable[[], _T],
    name: Optional[str] = None,
    field_name: Optional[str] = None,
    is_subscription: bool = False,
    description: Optional[str] = None,
    init: Literal[False] = False,
    permission_classes: Optional[List[Type[BasePermission]]] = None,
    deprecation_reason: Optional[str] = None,
    default: Any = dataclasses.MISSING,
    default_factory: Union[Callable[..., object], object] = dataclasses.MISSING,
    metadata: Optional[Mapping[Any, Any]] = None,
    directives: Optional[Sequence[object]] = (),
    pagination: Optional[bool] = UNSET,
    filters: Optional[type] = UNSET,
    order: Optional[type] = UNSET,
    only: Optional[TypeOrSequence[str]] = None,
    select_related: Optional[TypeOrSequence[str]] = None,
    prefetch_related: Optional[TypeOrSequence[PrefetchType]] = None,
    disable_optimization: bool = False,
) -> _T:
    ...


@overload
def field(
    *,
    name: Optional[str] = None,
    field_name: Optional[str] = None,
    is_subscription: bool = False,
    description: Optional[str] = None,
    init: Literal[True] = True,
    permission_classes: Optional[List[Type[BasePermission]]] = None,
    deprecation_reason: Optional[str] = None,
    default: Any = dataclasses.MISSING,
    default_factory: Union[Callable[..., object], object] = dataclasses.MISSING,
    metadata: Optional[Mapping[Any, Any]] = None,
    directives: Optional[Sequence[object]] = (),
    pagination: Optional[bool] = UNSET,
    filters: Optional[type] = UNSET,
    order: Optional[type] = UNSET,
    only: Optional[TypeOrSequence[str]] = None,
    select_related: Optional[TypeOrSequence[str]] = None,
    prefetch_related: Optional[TypeOrSequence[PrefetchType]] = None,
    disable_optimization: bool = False,
) -> Any:
    ...


@overload
def field(
    resolver: Union[StrawberryResolver, Callable, staticmethod, classmethod],
    *,
    name: Optional[str] = None,
    field_name: Optional[str] = None,
    is_subscription: bool = False,
    description: Optional[str] = None,
    permission_classes: Optional[List[Type[BasePermission]]] = None,
    deprecation_reason: Optional[str] = None,
    default: Any = dataclasses.MISSING,
    default_factory: Union[Callable[..., object], object] = dataclasses.MISSING,
    metadata: Optional[Mapping[Any, Any]] = None,
    directives: Optional[Sequence[object]] = (),
    pagination: Optional[bool] = UNSET,
    filters: Optional[type] = UNSET,
    order: Optional[type] = UNSET,
    only: Optional[TypeOrSequence[str]] = None,
    select_related: Optional[TypeOrSequence[str]] = None,
    prefetch_related: Optional[TypeOrSequence[PrefetchType]] = None,
    disable_optimization: bool = False,
) -> StrawberryDjangoField:
    ...


def field(
    resolver=None,
    *,
    name: Optional[str] = None,
    field_name: Optional[str] = None,
    is_subscription: bool = False,
    description: Optional[str] = None,
    permission_classes: Optional[List[Type[BasePermission]]] = None,
    deprecation_reason: Optional[str] = None,
    default: Any = dataclasses.MISSING,
    default_factory: Union[Callable[..., object], object] = dataclasses.MISSING,
    metadata: Optional[Mapping[Any, Any]] = None,
    directives: Optional[Sequence[object]] = (),
    pagination: Optional[bool] = UNSET,
    filters: Optional[type] = UNSET,
    order: Optional[type] = UNSET,
    only: Optional[TypeOrSequence[str]] = None,
    select_related: Optional[TypeOrSequence[str]] = None,
    prefetch_related: Optional[TypeOrSequence[PrefetchType]] = None,
    disable_optimization: bool = False,
    # This init parameter is used by pyright to determine whether this field
    # is added in the constructor or not. It is not used to change
    # any behavior at the moment.
    init: Literal[True, False, None] = None,
) -> Any:
    """Annotate a method or property as a Django GraphQL field.

    Examples:
        It can be used both as decorator and as a normal function:

        >>> @gql.type
        >>> class X:
        ...     field_abc: str = gql.django.field(description="ABC")
        ...     @gql.django.field(description="ABC")
        ...
        ...     def field_with_resolver(self) -> str:
        ...         return "abc"

    """
    f = StrawberryDjangoField(
        python_name=None,
        django_name=field_name,
        graphql_name=name,
        type_annotation=None,
        description=description,
        is_subscription=is_subscription,
        permission_classes=permission_classes or [],
        deprecation_reason=deprecation_reason,
        default=default,
        default_factory=default_factory,
        metadata=metadata,
        directives=directives,
        filters=filters,
        pagination=pagination,
        order=order,
        only=only,
        select_related=select_related,
        prefetch_related=prefetch_related,
        disable_optimization=disable_optimization,
    )
    if resolver:
        f = f(resolver)
    return f


def node(
    *,
    name: Optional[str] = None,
    is_subscription: bool = False,
    description: Optional[str] = None,
    permission_classes: Optional[List[Type[BasePermission]]] = None,
    deprecation_reason: Optional[str] = None,
    default: Any = dataclasses.MISSING,
    default_factory: Union[Callable[..., object], object] = dataclasses.MISSING,
    metadata: Optional[Mapping[Any, Any]] = None,
    directives: Optional[Sequence[object]] = (),
    only: Optional[TypeOrSequence[str]] = None,
    select_related: Optional[TypeOrSequence[str]] = None,
    prefetch_related: Optional[TypeOrSequence[PrefetchType]] = None,
    disable_optimization: bool = False,
    # This init parameter is used by pyright to determine whether this field
    # is added in the constructor or not. It is not used to change
    # any behavior at the moment.
    init: Literal[True, False, None] = None,
) -> Any:
    """Annotate a property to create a relay query field.

    Examples:
        Annotating something like this:

        >>> @strawberry.type
        >>> class X:
        ...     some_node: SomeType = relay.node(description="ABC")

        Will produce a query like this that returns `SomeType` given its id.

        ```
        query {
            someNode (id: ID) {
                id
                ...
            }
        }
        ```

    """
    return StrawberryDjangoNodeField(
        python_name=None,
        graphql_name=name,
        type_annotation=None,
        description=description,
        is_subscription=is_subscription,
        permission_classes=permission_classes or [],
        deprecation_reason=deprecation_reason,
        default=default,
        default_factory=default_factory,
        metadata=metadata,
        directives=directives or (),
        only=only,
        select_related=select_related,
        prefetch_related=prefetch_related,
        disable_optimization=disable_optimization,
    )


@overload
def connection(
    *,
    resolver: Callable[[], _T],
    name: Optional[str] = None,
    is_subscription: bool = False,
    description: Optional[str] = None,
    init: Literal[False] = False,
    permission_classes: Optional[List[Type[BasePermission]]] = None,
    deprecation_reason: Optional[str] = None,
    default: Any = dataclasses.MISSING,
    default_factory: Union[Callable[..., object], object] = dataclasses.MISSING,
    metadata: Optional[Mapping[Any, Any]] = None,
    directives: Optional[Sequence[object]] = (),
    filters: Optional[type] = UNSET,
    order: Optional[type] = UNSET,
    only: Optional[TypeOrSequence[str]] = None,
    select_related: Optional[TypeOrSequence[str]] = None,
    prefetch_related: Optional[TypeOrSequence[PrefetchType]] = None,
    disable_optimization: bool = False,
) -> _T:
    ...


@overload
def connection(
    *,
    name: Optional[str] = None,
    is_subscription: bool = False,
    description: Optional[str] = None,
    init: Literal[True] = True,
    permission_classes: Optional[List[Type[BasePermission]]] = None,
    deprecation_reason: Optional[str] = None,
    default: Any = dataclasses.MISSING,
    default_factory: Union[Callable[..., object], object] = dataclasses.MISSING,
    metadata: Optional[Mapping[Any, Any]] = None,
    directives: Optional[Sequence[object]] = (),
    filters: Optional[type] = UNSET,
    order: Optional[type] = UNSET,
    only: Optional[TypeOrSequence[str]] = None,
    select_related: Optional[TypeOrSequence[str]] = None,
    prefetch_related: Optional[TypeOrSequence[PrefetchType]] = None,
    disable_optimization: bool = False,
) -> Any:
    ...


@overload
def connection(
    resolver: Union[StrawberryResolver, Callable, staticmethod, classmethod],
    *,
    name: Optional[str] = None,
    is_subscription: bool = False,
    description: Optional[str] = None,
    permission_classes: Optional[List[Type[BasePermission]]] = None,
    deprecation_reason: Optional[str] = None,
    default: Any = dataclasses.MISSING,
    default_factory: Union[Callable[..., object], object] = dataclasses.MISSING,
    metadata: Optional[Mapping[Any, Any]] = None,
    directives: Optional[Sequence[object]] = (),
    filters: Optional[type] = UNSET,
    order: Optional[type] = UNSET,
    only: Optional[TypeOrSequence[str]] = None,
    select_related: Optional[TypeOrSequence[str]] = None,
    prefetch_related: Optional[TypeOrSequence[PrefetchType]] = None,
    disable_optimization: bool = False,
) -> StrawberryDjangoConnectionField:
    ...


def connection(
    resolver=None,
    *,
    name: Optional[str] = None,
    is_subscription: bool = False,
    description: Optional[str] = None,
    permission_classes: Optional[List[Type[BasePermission]]] = None,
    deprecation_reason: Optional[str] = None,
    default: Any = dataclasses.MISSING,
    default_factory: Union[Callable[..., object], object] = dataclasses.MISSING,
    metadata: Optional[Mapping[Any, Any]] = None,
    directives: Optional[Sequence[object]] = (),
    filters: Optional[type] = UNSET,
    order: Optional[type] = UNSET,
    only: Optional[TypeOrSequence[str]] = None,
    select_related: Optional[TypeOrSequence[str]] = None,
    prefetch_related: Optional[TypeOrSequence[PrefetchType]] = None,
    disable_optimization: bool = False,
    # This init parameter is used by pyright to determine whether this field
    # is added in the constructor or not. It is not used to change
    # any behavior at the moment.
    init: Literal[True, False, None] = None,
) -> Any:
    """Annotate a property or a method to create a relay connection field.

    Relay connections_ are mostly used for pagination purposes. This decorator
    helps creating a complete relay endpoint that provides default arguments
    and has a default implementation for the connection slicing.

    Note that when setting a resolver to this field, it is expected for this
    resolver to return an iterable of the expected node type, not the connection
    itself. That iterable will then be paginated accordingly. So, the main use
    case for this is to provide a filtered iterable of nodes by using some custom
    filter arguments.

    Examples:
        Annotating something like this:

        >>> @strawberry.type
        >>> class X:
        ...     some_node: relay.Connection[SomeType] = relay.connection(description="ABC")
        ...
        ...     @relay.connection(description="ABC")
        ...     def get_some_nodes(self, age: int) -> Iterable[SomeType]:
        ...         ...

        Will produce a query like this:

        ```
        query {
            someNode (before: String, after: String, first: String, after: String, age: Int) {
                totalCount
                pageInfo {
                    hasNextPage
                    hasPreviousPage
                    startCursor
                    endCursor
                }
                edges {
                    cursor
                    node {
                        id
                        ...
                    }
                }
            }
        }
        ```

    .. _Relay connections:
        https://relay.dev/graphql/connections.htm

    """
    f = StrawberryDjangoConnectionField(
        python_name=None,
        graphql_name=name,
        type_annotation=None,
        description=description,
        is_subscription=is_subscription,
        permission_classes=permission_classes or [],
        deprecation_reason=deprecation_reason,
        default=default,
        default_factory=default_factory,
        metadata=metadata,
        directives=directives or (),
        filters=filters,
        order=order,
        only=only,
        select_related=select_related,
        prefetch_related=prefetch_related,
        disable_optimization=disable_optimization,
    )
    if resolver is not None:
        f = f(resolver)
    return f



import contextvars
import dataclasses
import enum
import functools
from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    Iterable,
    List,
    Literal,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
    overload,
)

from django.core.exceptions import PermissionDenied
from django.db import models
from django.db.models import Model, QuerySet
from graphql.type.definition import GraphQLResolveInfo
import strawberry
from strawberry.django.context import StrawberryDjangoContext
from strawberry.private import Private
from strawberry.schema_directive import Location
from strawberry.types.info import Info
from strawberry.utils.await_maybe import AwaitableOrValue
from typing_extensions import Self, assert_never, final

from .directives import SchemaDirectiveHelper, SchemaDirectiveWithResolver
from .relay import Connection, GlobalID
from .types import OperationInfo, OperationMessage
from .utils import aio, resolvers
from .utils.query import filter_for_user
from .utils.typing import UserType

try:
    # Try to use the smaller/faster cache decorator if available
    _cache = functools.cache  # type:ignore
except AttributeError:
    _cache = functools.lru_cache

try:
    from .integrations.guardian import get_user_or_anonymous

    get_user_or_anonymous = resolvers.async_safe(get_user_or_anonymous)
except ImportError:
    # Access the user's id to force it to be loaded from the database
    get_user_or_anonymous = resolvers.async_safe(lambda u: (u, u.id)[0])


_T = TypeVar("_T")
_M = TypeVar("_M", bound=Model)

_user_ensured_attr = "_user_ensured"
_perm_safe_marker = "_strawberry_django_perm_safe_marker"

_return_condition = """\
When the condition fails, the following can happen (following this priority):
1) If the return value can return an `OperationMessage`, it is returned as that type.
2) If the field is mandatory (e.g. `String!`), this will result in an error.
3) If the field is not mandatory and a list (e.g. `[String]`), an empty list will be returned.
4) If the field is not mandatory and any scalar or object (e.g. `String`), `null` will be returned.
5) If the field is a relay `Connection`, an empty connection will be returned.
"""
_desc = lambda desc: f"{desc}\n\n{_return_condition.strip()}"

perm_safe: contextvars.ContextVar[Optional[List[bool]]] = contextvars.ContextVar(
    "perm_safe",
    default=None,
)
running_checks: contextvars.ContextVar[List["HasPermDirective"]] = contextvars.ContextVar(
    "running_checks",
    default=[],
)


def init_checker(checker: "HasPermDirective"):
    checks = running_checks.get()[:]
    checks.append(checker)
    running_checks.set(checks)

    perms = perm_safe.get() or []
    perm_safe.set(perms[:])


def clear_checker():
    perm_safe.set(None)
    checks = running_checks.get()
    if checks:
        checks.clear()


def set_perm_safe(value: bool):
    ps = perm_safe.get()
    if ps is not None:
        ps.append(value)


def is_perm_safe():
    ps = perm_safe.get()
    return bool(ps and all(ps))


def filter_with_perms(qs: QuerySet[_M], info: Info) -> QuerySet[_M]:
    checks = running_checks.get()
    if not checks:
        return qs

    # Do not do anything is results are cached, the target is the the retval
    if qs._result_cache is not None:  # type:ignore
        set_perm_safe(False)
        return qs

    user = cast(StrawberryDjangoContext, info.context).request.user
    # If the user is anonymous, we can't filter object permissions for it
    if user.is_anonymous:
        set_perm_safe(False)
        return qs.none()

    for check in checks:
        if check.target != PermTarget.RETVAL:
            continue

        qs = filter_for_user(
            qs,
            cast(UserType, user),
            [p.perm for p in check.permissions],
            any_perm=check.any,
            with_superuser=check.with_superuser,
        )

    set_perm_safe(True)
    return qs


@overload
def get_with_perms(
    pk: strawberry.ID,
    info: Info,
    *,
    required: Literal[True],
    model: Type[_M],
) -> _M:
    ...


@overload
def get_with_perms(
    pk: strawberry.ID,
    info: Info,
    *,
    required: bool = ...,
    model: Type[_M],
) -> Optional[_M]:
    ...


@overload
def get_with_perms(
    pk: GlobalID,
    info: Info,
    *,
    required: Literal[True],
    model: Type[_M],
) -> _M:
    ...


@overload
def get_with_perms(
    pk: GlobalID,
    info: Info,
    *,
    required: bool = ...,
    model: Type[_M],
) -> Optional[_M]:
    ...


@overload
def get_with_perms(
    pk: GlobalID,
    info: Info,
    *,
    required: Literal[True],
) -> Any:
    ...


@overload
def get_with_perms(
    pk: GlobalID,
    info: Info,
    *,
    required: bool = ...,
) -> Optional[Any]:
    ...


def get_with_perms(pk, info, *, required=False, model=None):
    if isinstance(pk, GlobalID):
        instance = pk.resolve_node(info, required=required, ensure_type=model)
        if aio.is_awaitable(instance, info=info):
            instance = resolvers.resolve_sync(instance)
        instance = cast(models.Model, instance)
    else:
        assert model
        instance = model._default_manager.get(pk=pk)

    if instance is None:
        return None

    checks = running_checks.get()
    if not checks:
        return instance

    user = cast(StrawberryDjangoContext, info.context).request.user
    for check in checks:
        f = any if check.any else all
        checker = check.obj_perm_checker(info, cast(UserType, user))
        if not f(checker(p, instance) for p in check.permissions):
            raise PermissionDenied(check.message)

    return instance


@dataclasses.dataclass
class AuthDirective(SchemaDirectiveWithResolver):
    """Base auth directive definition."""

    has_resolver: ClassVar = True

    @property
    def message(self) -> str:
        return "User is not authorized."

    def resolve(
        self,
        helper: SchemaDirectiveHelper,
        _next: Callable,
        root: Any,
        info: GraphQLResolveInfo,
        *args,
        **kwargs,
    ):
        context = cast(StrawberryDjangoContext, info.context)
        resolver = functools.partial(_next, root, info, *args, **kwargs)

        user = cast(UserType, context.request.user)
        if not getattr(context, _user_ensured_attr, False):
            return aio.resolve(
                cast(UserType, get_user_or_anonymous(user)),
                functools.partial(
                    self.resolve_for_user,
                    helper,
                    resolver,
                    root,
                    info,
                    **kwargs,
                ),
                info=info,
            )

        return self.resolve_for_user(
            helper,
            resolver,
            root,
            info,
            cast(UserType, user),
            **kwargs,
        )

    def resolve_for_user(
        self,
        helper: SchemaDirectiveHelper,
        resolver: Callable,
        root: Any,
        info: GraphQLResolveInfo,
        user: UserType,
        **kwargs,
    ):
        raise NotImplementedError

    def resolve_retval(
        self,
        helper: SchemaDirectiveHelper,
        root: Any,
        info: GraphQLResolveInfo,
        retval: Any,
        auth_ok: AwaitableOrValue[bool],
    ):
        # If this is not bool, assume async. Avoid is_awaitable since it is slow
        if not isinstance(auth_ok, bool):
            return aio.resolve_async(
                auth_ok,
                functools.partial(self.resolve_retval, helper, root, info, retval),
            )

        # Make sure any chained resolvers will not try to validate the result again
        clear_checker()

        if auth_ok:
            if callable(retval):
                retval = retval()
            return retval

        for p in helper.ret_possibilities:
            if p.type_def and issubclass(p.type_def.origin, OperationMessage):
                return p.type_def.origin(
                    kind=OperationMessage.Kind.PERMISSION,
                    message=self.message,
                    field=info.field_name,
                )
            elif p.type_def and issubclass(p.type_def.origin, OperationInfo):
                return p.type_def.origin(
                    messages=[
                        OperationMessage(
                            kind=OperationMessage.Kind.PERMISSION,
                            message=self.message,
                            field=info.field_name,
                        )
                    ],
                )

        # If the field is optional, return null
        if helper.optional:
            return None

        # If it is a list, return an empty list
        if helper.is_list:
            return []

        # If it is a Connection, try to return an empty connection, but only if
        # it is the only possibility available...
        if len(helper.ret_possibilities) == 1:
            type_def = helper.ret_possibilities[0].type_def
            if (
                type_def
                and type_def.concrete_of
                and issubclass(type_def.concrete_of.origin, Connection)
            ):
                return type_def.origin.from_nodes([], total_count=0)

        # In last case, raise an error
        raise PermissionDenied(self.message)


@dataclasses.dataclass
class ConditionDirective(AuthDirective):
    """Base directive for condition checking."""

    priority: ClassVar = 99
    message: Private[str] = dataclasses.field(default="User does not have permission.")

    def __init_subclass__(cls) -> None:
        for attr in ["__hash__", "__eq__"]:
            if attr not in cls.__dict__:
                setattr(cls, attr, getattr(cls, attr))
        return super().__init_subclass__()

    def __hash__(self):
        return hash(self.__class__)

    def __eq__(self, other: Self):
        return self.__class__ == other.__class__

    def resolve_for_user(
        self,
        helper: SchemaDirectiveHelper,
        resolver: Callable,
        root: Any,
        info: GraphQLResolveInfo,
        user: UserType,
        **kwargs,
    ):
        return self.resolve_retval(
            helper,
            root,
            info,
            resolver,
            self.check_condition(root, info, user, **kwargs),
        )

    def check_condition(
        self, root: Any, info: GraphQLResolveInfo, user: UserType, **kwargs
    ) -> bool:
        raise NotImplementedError


@strawberry.schema_directive(
    locations=[Location.FIELD_DEFINITION],
    description=_desc("Can only be resolved by authenticated users."),
)
@final
class IsAuthenticated(ConditionDirective):
    """Mark a field as only resolvable by authenticated users."""

    message: Private[str] = dataclasses.field(default="User is not authenticated.")

    def check_condition(
        self, root: Any, info: GraphQLResolveInfo, user: UserType, **kwargs
    ) -> bool:
        return user.is_authenticated and user.is_active


@strawberry.schema_directive(
    locations=[Location.FIELD_DEFINITION],
    description=_desc("Can only be resolved by staff users."),
)
@final
class IsStaff(ConditionDirective):
    """Mark a field as only resolvable by staff users."""

    message: Private[str] = dataclasses.field(default="User is not a staff member.")

    def check_condition(
        self, root: Any, info: GraphQLResolveInfo, user: UserType, **kwargs
    ) -> bool:
        return user.is_authenticated and user.is_staff


@strawberry.schema_directive(
    locations=[Location.FIELD_DEFINITION],
    description=_desc("Can only be resolved by superuser users."),
)
@final
class IsSuperuser(ConditionDirective):
    """Mark a field as only resolvable by superuser users."""

    message: Private[str] = dataclasses.field(default="User is not a superuser.")

    def check_condition(
        self, root: Any, info: GraphQLResolveInfo, user: UserType, **kwargs
    ) -> bool:
        return user.is_authenticated and user.is_superuser


@strawberry.input(description="Permission definition for schema directives.")
@dataclasses.dataclass(eq=True, order=True, frozen=True)
class PermDefinition:
    """Permission definition.

    Attributes:
        resource:
            The resource to which we are requiring permission.
        permission:
            The permission itself

    """

    resource: Optional[str] = strawberry.field(
        description=(
            "The resource to which we are requiring permission. If this is "
            "empty that means that we are checking the permission directly."
        ),
    )
    permission: Optional[str] = strawberry.field(
        description=(
            "The permission itself. If this is empty that means that we "
            "are checking for any permission for the given resource."
        ),
    )

    @classmethod
    def from_perm(cls, perm: str):
        parts = perm.split(".")
        if len(parts) != 2:
            raise TypeError(
                "Permissions need to be defined as `app_label.perm`, `app_label.` or `.perm`"
            )
        return cls(
            resource=parts[0].strip() or None,
            permission=parts[1].strip() or None,
        )

    @property
    def perm(self):
        return f"{self.resource or ''}.{self.permission or ''}".strip(".")


class PermTarget(enum.Enum):
    """Permission location."""

    ROOT = "root"
    RETVAL = "retval"


@dataclasses.dataclass
class HasPermDirective(AuthDirective):
    """Permission directive."""

    target: ClassVar[Optional[PermTarget]]

    # FIXME: This is to allow passing a str or list[str] to perms, but still print
    perms: strawberry.Private[Union[List[str], str]]

    permissions: List[PermDefinition] = strawberry.field(
        description="Required perms to access this resource.",
        default_factory=list,
    )
    any: bool = strawberry.field(  # noqa:A003
        description="If any or all perms listed are required.",
        default=True,
    )
    message: Private[str] = dataclasses.field(
        default="You don't have permission to access this resource.",
    )
    perm_checker: Private[
        Callable[[GraphQLResolveInfo, UserType], Callable[[PermDefinition], bool]]
    ] = dataclasses.field(
        default=lambda info, user: lambda perm: (
            user.has_perm(perm.perm)
            if perm.permission
            else user.has_module_perms(cast(str, perm.resource))
        ),
    )
    obj_perm_checker: Private[
        Callable[[GraphQLResolveInfo, UserType], Callable[[PermDefinition, Any], bool]]
    ] = dataclasses.field(
        default=lambda info, user: lambda perm, obj: (
            # Check global perms first, then object specific
            user.has_perm(perm.perm)
            or user.has_perm(perm.perm, obj=obj)
        ),
    )
    with_anonymous: Private[bool] = dataclasses.field(default=True)
    with_superuser: Private[bool] = dataclasses.field(default=False)

    def __post_init__(self):
        super().__post_init__()

        perms = self.perms
        if isinstance(perms, str):
            perms = [perms]
        if isinstance(perms, Iterable):
            perms = [PermDefinition.from_perm(p) if isinstance(p, str) else p for p in perms]

        if not len(self.perms):
            raise TypeError(f"At least one perm is required for {self!r}")

        assert all(isinstance(p, PermDefinition) for p in perms)
        self.permissions = perms

    def __init_subclass__(cls) -> None:
        for attr in ["__hash__", "__eq__"]:
            if attr not in cls.__dict__:
                setattr(cls, attr, getattr(cls, attr))
        return super().__init_subclass__()

    def __hash__(self):
        return hash(
            (
                self.target,
                frozenset(self.perms),
                self.any,
                self.perm_checker,
                self.obj_perm_checker,
            )
        )

    def __eq__(self, other: Self):
        return (
            self.__class__ == other.__class__
            and self.target == other.target
            and set(self.perms) == set(other.perms)
            and self.any == other.any
            and self.message == other.message
            and self.perm_checker == other.perm_checker
            and self.obj_perm_checker == other.obj_perm_checker
            and self.with_anonymous == other.with_anonymous
            and self.with_superuser == other.with_superuser
        )

    def get_cache(
        self,
        info: GraphQLResolveInfo,
        user: UserType,
    ) -> Dict[Union[Self, Tuple[Self, Any]], bool]:
        cache_key = f"_{self.__class__.__name__}_cache"

        cache = getattr(user, cache_key, None)
        if cache is not None:
            return cache

        cache = {}
        setattr(user, cache_key, cache)
        return cache

    def resolve_for_user(
        self,
        helper: SchemaDirectiveHelper,
        resolver: Callable,
        root: Any,
        info: GraphQLResolveInfo,
        user: UserType,
        **kwargs,
    ):
        if self.with_superuser and user.is_active and user.is_superuser:
            return self.resolve_retval(helper, root, info, resolver, True)
        if self.with_anonymous and user.is_anonymous:
            return self.resolve_retval(helper, root, info, resolver, False)

        cache = self.get_cache(info, user)

        if self.target is None:
            has_perm = cache.get(self)
            if has_perm is None:
                has_perm = self._has_perm_safe(root, info, user)
            return self.resolve_retval(helper, root, info, resolver, has_perm)
        elif self.target == PermTarget.ROOT:
            has_perm = cache.get((self, root))
            if has_perm is None:
                has_perm = self._has_obj_perm_safe(root, info, user, root)
            return self.resolve_retval(helper, root, info, resolver, has_perm)
        elif self.target == PermTarget.RETVAL:
            init_checker(self)

            ret = resolver()
            if ret is None:
                # Retval is None, just return it
                return None

            # Avoid is_awaitable as much as we can
            if not isinstance(ret, (list, Model, QuerySet)) and aio.is_awaitable(ret, info=info):
                return aio.resolve_async(
                    ret,
                    functools.partial(self._resolve_obj_perms, helper, root, info, user),
                )
            return self._resolve_obj_perms(helper, root, info, user, ret)
        else:
            assert_never(self.target)  # noqa:R503

    @resolvers.async_safe
    def _has_perm_safe(
        self,
        root: Any,
        info: GraphQLResolveInfo,
        user: UserType,
    ) -> bool:
        cache = self.get_cache(info, user)

        # Maybe the result ended up in the cache in the meantime
        if self in cache:
            return cache[self]

        f = any if self.any else all
        checker = self.perm_checker(info, user)
        has_perm = f(checker(p) for p in self.permissions)
        cache[self] = has_perm

        return has_perm

    @resolvers.async_safe
    def _has_obj_perm_safe(
        self,
        root: Any,
        info: GraphQLResolveInfo,
        user: UserType,
        obj: Any,
    ) -> bool:
        cache = self.get_cache(info, user)

        # Maybe the result ended up in the cache in the meantime
        key = (self, obj)
        if key in cache:
            return cache[key]

        f = any if self.any else all
        checker = self.obj_perm_checker(info, user)
        has_perm = f(checker(p, obj) for p in self.permissions)

        cache[key] = has_perm
        return has_perm

    def _resolve_obj_perms(
        self,
        helper: SchemaDirectiveHelper,
        root: Any,
        info: GraphQLResolveInfo,
        user: UserType,
        obj: Any,
    ) -> Any:
        if is_perm_safe():
            return self.resolve_retval(helper, root, info, obj, True)

        if isinstance(obj, Iterable):
            return self._resolve_iterable_perms_safe(helper, root, info, user, obj)

        cache = self.get_cache(info, user)
        has_perm = cache.get((self, obj))
        if has_perm is None:
            has_perm = self._has_obj_perm_safe(root, info, user, obj)

        return self.resolve_retval(helper, root, info, obj, has_perm)

    @resolvers.async_safe
    def _resolve_iterable_perms_safe(
        self,
        helper: SchemaDirectiveHelper,
        root: Any,
        info: GraphQLResolveInfo,
        user: UserType,
        objs: Iterable[Any],
    ) -> Any:
        if is_perm_safe():
            return self.resolve_retval(helper, root, info, objs, True)

        cache = self.get_cache(info, user)
        f = any if self.any else all
        checker = self.obj_perm_checker(info, user)

        def _check_obj(obj):
            key = (self, obj)
            if key in cache:
                return cache[key]

            has_perm = f(checker(p, obj) for p in self.permissions)
            cache[key] = has_perm

            return has_perm

        objs = [obj for obj in objs if _check_obj(obj)]
        return self.resolve_retval(helper, root, info, objs, True)


@strawberry.schema_directive(
    locations=[Location.FIELD_DEFINITION],
    description=_desc("Will check if the user has any/all permissions to resolve this."),
)
@final
class HasPerm(HasPermDirective):
    """Defines permissions required to access the given object/field.

    Given a `resource` name, the user can access the decorated object/field
    if he has any of the permissions defined in this directive.

    Examples:
        To indicate that a mutation can only be done by someone who
        has "product.add_product" perm in the django system:

        >>> @strawberry.type
        ... class Query:
        ...     @strawberry.mutation(directives=[PermRequired("product.add_product")])
        ...     def create_product(self, name: str) -> ProductType:
        ...         ...

    Attributes:
        perms:
            Perms required to access this resource.
        any:
            If any perm or all perms are required to resolve the object/field.
        with_anonymous:
            If we should optimize the permissions check and consider an anonymous
            user as not having any permissions. This is true by default, which means
            that anonymous users will not trigger has_perm checks.
        with_superuser:
            If we should optimize the permissions check and consider a superuser
            as having permissions foe everything. This is false by default to avoid
            returning unexpected results. Setting this to true will avoid triggering
            has_perm checks.

    """

    priority: ClassVar = 55
    target: ClassVar = None


@strawberry.schema_directive(
    locations=[Location.FIELD_DEFINITION],
    description=_desc(
        "Will check if the user has any/all permissions for the parent "
        "of this field to resolve this."
    ),
)
@final
class HasRootPerm(HasPermDirective):
    """Defines permissions required to access the given field at object level.

    This will check the permissions for the root object to access the given field.

    Unlike `ObjPermRequired`, this uses the root value (the object where the field
    is defined) to resolve the field, which means that this cannot be used for root
    queries and types.

    Examples:
        To indicate that a field inside a `ProductType` can only be accessed if
        the user has "product.view_field" in it in the django system:

        >>> @gql.django.type(Product)
        ... class ProdyctType:
        ...     some_field: str = strawberry.field(
        ...         directives=[RootPermRequired(".add_product")],
        ...     )

    Attributes:
        perms:
            Perms required to access this resource.
        any:
            If any perm or all perms are required to resolve the object/field.
        checker:
            An optional callable to check if the user can access that object.
            By default it resolves using `user.has_perm(perm, retval)`
        with_anonymous:
            If we should optimize the permissions check and consider an anonymous
            user as not having any permissions. This is true by default, which means
            that anonymous users will not trigger has_perm checks.
        with_superuser:
            If we should optimize the permissions check and consider a superuser
            as having permissions foe everything. This is false by default to avoid
            returning unexpected results. Setting this to true will avoid triggering
            has_perm checks.

    """

    priority: ClassVar = 50
    target: ClassVar = PermTarget.ROOT


@strawberry.schema_directive(
    locations=[Location.FIELD_DEFINITION],
    description=_desc(
        "Will check if the user has any/all permissions for the resolved "
        "value of this field before returning it."
    ),
)
@final
class HasObjPerm(HasPermDirective):
    """Defines permissions required to access the given object/field at object level.

    Given a `resource` name, the user can access the decorated object/field
    if he has any of the permissions defined in this directive.

    Note that this depends on resolving the object to check the permissions
    specifically for that object, unlike `PermRequired` which checks it before resolving.

    Examples:
        To indicate that a field that returns a `ProductType` can only be accessed
        by someone who has "product.view_product"
        has "product.view_product" perm in the django system:

        >>> @strawberry.type
        ... class SomeType:
        ...     product: ProductType = strawberry.field(
        ...         directives=[ObjPermRequired(".add_product")],
        ...     )

    Attributes:
        perms:
            Perms required to access this resource.
        any:
            If any perm or all perms are required to resolve the object/field.
        checker:
            An optional callable to check if the user can access that object.
            By default it resolves using `user.has_perm(perm, retval)`
        with_anonymous:
            If we should optimize the permissions check and consider an anonymous
            user as not having any permissions. This is true by default, which means
            that anonymous users will not trigger has_perm checks.
        with_superuser:
            If we should optimize the permissions check and consider a superuser
            as having permissions foe everything. This is false by default to avoid
            returning unexpected results. Setting this to true will avoid triggering
            has_perm checks.
            Note that lists and relay connections filter the results based on the
            existence of the exact permissions for the user, which means that when
            this is true, no results would be filtered, which might not be the
            expected result.

    """

    target: ClassVar = PermTarget.RETVAL



import dataclasses
from typing import TYPE_CHECKING, Final, List, Optional

from django.conf import settings

__all__ = [
    "NAMESPACE",
    "Config",
    "config",
]

NAMESPACE: Final = "STRAWBERRY_DJANGO"


@dataclasses.dataclass
class Config:
    """Settings for this extension.

    The configs here are expected to be set on django like:

        "STRAWBERRY_DJANGO_<ATTNAME>"

    """

    RELAY_MAX_RESULTS: Optional[int] = dataclasses.field(default=100)
    REMOVE_DUPLICATED_SUFFIX: List[str] = dataclasses.field(
        default_factory=lambda: ["Input", "Partial"],
    )
    FIELDS_USE_GLOBAL_ID: bool = dataclasses.field(default=True)

    # Trick type checking into thinking that we only have the defined configs
    if not TYPE_CHECKING:

        def __getattribute__(self, attr: str):
            config_name = f"{NAMESPACE}_{attr}"
            if hasattr(settings, config_name):
                return getattr(settings, config_name)
            return object.__getattribute__(self, attr)


config: Final = Config()



from strawberry_django import (
    DjangoFileType,
    DjangoImageType,
    DjangoModelType,
    ManyToManyInput,
    ManyToOneInput,
    OneToManyInput,
    OneToOneInput,
    auth,
    django_resolver,
    filters,
    ordering,
)

from strawberry_django_plus import mutations, types
from strawberry_django_plus.field import connection, field, node
from strawberry_django_plus.filters import filter
from strawberry_django_plus.mutations.fields import create as create_mutation
from strawberry_django_plus.mutations.fields import delete as delete_mutation
from strawberry_django_plus.mutations.fields import input_mutation, mutation
from strawberry_django_plus.mutations.fields import update as update_mutation
from strawberry_django_plus.ordering import order
from strawberry_django_plus.type import input, interface, partial, type

__all__ = [
    # strawberry_django
    "auth",
    "django_resolver",
    "filters",
    "ordering",
    "DjangoFileType",
    "DjangoImageType",
    "DjangoModelType",
    "OneToOneInput",
    "OneToManyInput",
    "ManyToOneInput",
    "ManyToManyInput",
    # strawberry_django_plus
    "create_mutation",
    "delete_mutation",
    "update_mutation",
    "field",
    "node",
    "connection",
    "input",
    "mutation",
    "input_mutation",
    "interface",
    "mutations",
    "type",
    "types",
    "partial",
    "filter",
    "order",
]



# This should go to its own module or be contributed back to strawberry
import abc
import base64
import dataclasses
import functools
import inspect
import math
import sys
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Callable,
    ClassVar,
    Dict,
    ForwardRef,
    Generic,
    Iterable,
    List,
    Literal,
    Mapping,
    Optional,
    Sequence,
    Set,
    Sized,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
    get_args,
    get_origin,
    overload,
)
from typing import _eval_type  # type:ignore
import uuid

from graphql import GraphQLID
import strawberry
from strawberry.annotation import StrawberryAnnotation
from strawberry.arguments import StrawberryArgument
from strawberry.custom_scalar import ScalarDefinition
from strawberry.field import StrawberryField
from strawberry.lazy_type import LazyType
from strawberry.permission import BasePermission
from strawberry.schema.types.scalar import DEFAULT_SCALAR_REGISTRY
from strawberry.type import (
    StrawberryList,
    StrawberryOptional,
    StrawberryType,
    StrawberryTypeVar,
)
from strawberry.types import Info
from strawberry.types.fields.resolver import StrawberryResolver
from strawberry.types.types import TypeDefinition
from strawberry.utils.await_maybe import AwaitableOrValue
from strawberry.utils.str_converters import to_camel_case
from typing_extensions import Annotated

from .settings import config
from .utils import aio

__all__ = [
    "Connection",
    "ConnectionField",
    "Edge",
    "GlobalID",
    "GlobalIDValueError",
    "Node",
    "NodeField",
    "NodeType",
    "PageInfo",
    "RelayField",
    "from_base64",
    "to_base64",
    "node",
    "connection",
]

_T = TypeVar("_T")
_R = TypeVar("_R")
connection_typename = "arrayconnection"

NodeType = TypeVar("NodeType", bound="Node")


def from_base64(value: str) -> Tuple[str, str]:
    """Parse the base64 encoded relay value.

    Args:
        value:
            The value to be parsed

    Returns:
        A tuple of (TypeName, NodeID).

    Raises:
        ValueError:
            If the value is not in the expected format

    """
    try:
        res = base64.b64decode(value.encode()).decode().split(":", 1)
    except Exception as e:
        raise ValueError(str(e)) from e

    if len(res) != 2:
        raise ValueError(f"{res} expected to contain only 2 items")

    return res[0], res[1]


def to_base64(type_: Union[str, type, TypeDefinition], node_id: Any) -> str:
    """Encode the type name and node id to a base64 string.

    Args:
        type_:
            The GraphQL type, type definition or type name.
        node_id:
            The node id itself

    Returns:
        A tuple of (TypeName, NodeID).

    Raises:
        ValueError:
            If the value is not a valid GraphQL type or name

    """
    try:
        if isinstance(type_, str):
            type_name = type_
        elif isinstance(type_, TypeDefinition):
            type_name = type_.name
        elif isinstance(type_, type):
            type_name = type_._type_definition.name  # type:ignore
    except Exception as e:
        raise ValueError(f"{type_} is not a valid GraphQL type or name") from e

    return base64.b64encode(f"{type_name}:{node_id}".encode()).decode()


class GlobalIDValueError(ValueError):
    """GlobalID value error, usually related to parsing or serialization."""


@dataclasses.dataclass(order=True, frozen=True)
class GlobalID:
    """Global ID for relay types.

    Different from `strawberry.ID`, this ID wraps the original object ID in a string
    that contains both its GraphQL type name and the ID itself, and encodes it
    to a base64_ string.

    This object contains helpers to work with that, including method to retrieve
    the python object type or even the encoded node itself.

    Attributes:
        type_name:
            The type name part of the id
        node_id:
            The node id part of the id

    .. _base64:
        https://en.wikipedia.org/wiki/Base64

    """

    _nodes_cache: ClassVar[Dict[Tuple[int, str], Type["Node"]]] = {}

    type_name: str
    node_id: str

    def __post_init__(self):
        if not isinstance(self.type_name, str):
            raise GlobalIDValueError(
                f"type_name is expected to be a string, found {repr(self.type_name)}"
            )
        if not isinstance(self.node_id, str):
            raise GlobalIDValueError(
                f"node_id is expected to be a string, found {repr(self.node_id)}"
            )

    def __str__(self):
        return to_base64(self.type_name, self.node_id)

    @classmethod
    def from_id(cls, value: Union[str, strawberry.ID]):
        """Create a new GlobalID from parsing the given value.

        Args:
            value:
                The value to be parsed, as a base64 string in the "TypeName:NodeID" format

        Returns:
            An instance of GLobalID

        Raises:
            GlobalIDValueError:
                If the value is not in a GLobalID format

        """
        try:
            type_name, node_id = from_base64(value)
        except ValueError as e:
            raise GlobalIDValueError(str(e)) from e

        return cls(type_name=type_name, node_id=node_id)

    def resolve_type(self, info: Info) -> Type["Node"]:
        """Resolve the internal type name to its type itself.

        Args:
            info:
                The strawberry execution info resolve the type name from

        Returns:
            The resolved GraphQL type for the execution info

        """
        schema = info.schema
        # Put the schema in the key so that different schemas can have different types
        key = (id(schema), self.type_name)
        origin = self._nodes_cache.get(key)

        if origin is None:
            type_def = info.schema.get_type_by_name(self.type_name)
            assert isinstance(type_def, TypeDefinition)
            origin = type_def.origin
            if isinstance(origin, LazyType):
                origin = origin.resolve_type()
            assert issubclass(origin, Node)
            self._nodes_cache[key] = origin

        return origin

    @overload
    def resolve_node(
        self,
        info: Info,
        *,
        required: Literal[True] = ...,
        ensure_type: Type[_T],
    ) -> _T:
        ...

    @overload
    def resolve_node(
        self,
        info: Info,
        *,
        required: Literal[True] = ...,
        ensure_type: Awaitable[Type[_T]],
    ) -> Awaitable[_T]:
        ...

    @overload
    def resolve_node(
        self,
        info: Info,
        *,
        required: Literal[True],
        ensure_type: None = ...,
    ) -> AwaitableOrValue["Node"]:
        ...

    @overload
    def resolve_node(
        self,
        info: Info,
        *,
        required: bool = ...,
        ensure_type: None = ...,
    ) -> AwaitableOrValue[Optional["Node"]]:
        ...

    def resolve_node(self, info, *, required=False, ensure_type=None) -> Any:
        """Resolve the type name and node id info to the node itself.

        Tip: When you know the expected type, calling `ensure_type` should help
        not only to enforce it, but also help with typing since it will know that,
        if this function returns successfully, the retval should be of that
        type and not `Node`.

        Args:
            info:
                The strawberry execution info resolve the type name from
            required:
                If the value is required to exist. Note that asking to ensure
                the type automatically makes required true.
            ensure_type:
                Optionally check if the returned node is really an instance
                of this type.

        Returns:
            The resolved node

        Raises:
            TypeError:
                If ensure_type was provided and the type is not an instance of it

        """
        n_type = self.resolve_type(info)
        node = n_type.resolve_node(
            self.node_id,
            info=info,
            required=required or ensure_type is not None,
        )

        if node is not None and ensure_type is not None:
            origin = get_origin(ensure_type)
            if origin and issubclass(origin, Awaitable):
                ensure_type = get_args(ensure_type)[0]
            return aio.resolve(node, lambda n: n, info=info, ensure_type=ensure_type)

        return node


# Register our GlobalID scalar
DEFAULT_SCALAR_REGISTRY[GlobalID] = ScalarDefinition(
    # Use the same name/description/parse_literal from GraphQLID as relay
    # specs expect this type to be "ID".
    name="GlobalID",
    description=GraphQLID.description,
    parse_literal=lambda v, vars=None: GlobalID.from_id(GraphQLID.parse_literal(v, vars)),
    parse_value=GlobalID.from_id,
    serialize=str,
    specified_by_url="https://relay.dev/graphql/objectidentification.htm",
)


@strawberry.interface(description="An object with a Globally Unique ID")
class Node(abc.ABC):
    """Node interface for GraphQL types.

    All types that are relay ready should inherit from this interface and
    implement the following methods.

    Attributes:
        id_attr:
            (Optional) Define id field of node

    Methods:
        resolve_id:
            (Optional) Called to resolve the node's id.
            By default it returns `getattr(node, getattr(node, 'id_attr'. 'id'))`
            to use the one provided when creating the node itself.
        resolve_node:
            Called to retrieve a node given its id
        resolve_nodes:
            Called to retrieve an iterable of node given their ids
        resolve_connection:
            (Optional) Called to resolve a `Connection` to this node. Override
            this to modify how the connection is resolved.

    """

    # We can't do CONNECTION_CLASS = Connection because it is not defined yet.
    # We will define it below though as the default one.
    if TYPE_CHECKING:
        CONNECTION_CLASS: ClassVar[Type["Connection"]]

    @strawberry.field(description="The Globally Unique ID of this object")
    @classmethod
    def id(cls, root: "Node", info: Info) -> GlobalID:  # noqa:A003
        # FIXME: We want to support both integration objects that doesn't define a resolve_id
        # and also the ones that does override it. Is there a better way of handling this?
        if isinstance(root, Node):
            resolve_id = root.__class__.resolve_id
        else:
            resolve_id = cls.resolve_id

        node_id = resolve_id(root, info=info)
        resolve_typename = (
            root.__class__.resolve_typename if isinstance(root, Node) else cls.resolve_typename
        )
        type_name = resolve_typename(root, info)
        assert type_name

        if isinstance(node_id, str):
            # str is the default and is faster to check for it than is_awaitable
            return GlobalID(type_name=type_name, node_id=node_id)
        elif isinstance(node_id, (int, uuid.UUID)):
            # those are very common ids and are safe to convert to str
            return GlobalID(type_name=type_name, node_id=str(node_id))
        elif aio.is_awaitable(node_id, info=info):
            return aio.resolve_async(  # type:ignore
                node_id,
                lambda resolved: GlobalID(type_name=type_name, node_id=resolved),
                info=info,
            )

        # If node_id is not str, GlobalID will raise an error for us
        return GlobalID(type_name=type_name, node_id=cast(str, node_id))

    @classmethod
    def resolve_id(
        cls: Type[NodeType],
        root: NodeType,
        *,
        info: Optional[Info] = None,
    ) -> AwaitableOrValue[str]:
        """Resolve the node id.

        By default this returns `getattr(node, getattr(node, 'id_attr'. 'id'))`.
        Override this to return something else.

        Args:
            info:
                The strawberry execution info resolve the type name from
            root:
                The node to resolve

        Returns:
            The resolved id (which is expected to be str)

        """
        id_attr = getattr(cls, "id_attr", "id")
        return getattr(root, id_attr)

    @classmethod
    def resolve_typename(cls: Type[NodeType], root: NodeType, info: Info):
        return info.path.typename

    @classmethod
    def resolve_connection(
        cls: Type[NodeType],
        *,
        info: Optional[Info] = None,
        nodes: Optional[AwaitableOrValue[Iterable[NodeType]]] = None,
        total_count: Optional[int] = None,
        before: Optional[str] = None,
        after: Optional[str] = None,
        first: Optional[int] = None,
        last: Optional[int] = None,
    ) -> AwaitableOrValue["Connection[NodeType]"]:
        """Resolve a connection for this node.

        By default this will call `cls.resolve_nodes` if None were provided,
        and them the connection will be generated from `Connection.from_nodes`.

        Args:
            info:
                The strawberry execution info resolve the type name from
            nodes:
                An iterable of nodes to transform to a connection
            total_count:
                Optionally provide a total count so that the connection
                doesn't have to calculate it. Might be useful for some ORMs
                for performance reasons.
            before:
                Returns the items in the list that come before the specified cursor
            after:
                Returns the items in the list that come after the specified cursor
            first:
                Returns the first n items from the list
            last:
                Returns the items in the list that come after the specified cursor

        Returns:
            The resolved `Connection`

        """
        if nodes is None:
            nodes = cls.resolve_nodes(info=info)

        if aio.is_awaitable(nodes, info=info):
            return aio.resolve_async(
                nodes,
                lambda resolved: cls.resolve_connection(
                    info=info,
                    nodes=resolved,
                    total_count=total_count,
                    before=before,
                    after=after,
                    first=first,
                    last=last,
                ),
                info=info,
            )

        # FIXME: Remove cast once pyright resolves the negative TypeGuard form
        nodes = cast(Iterable[NodeType], nodes)

        return cls.CONNECTION_CLASS.from_nodes(
            nodes,
            total_count=total_count,
            before=before,
            after=after,
            first=first,
            last=last,
        )

    @classmethod
    @abc.abstractmethod
    def resolve_nodes(
        cls: Type[NodeType],
        *,
        info: Optional[Info] = None,
        node_ids: Optional[Iterable[str]] = None,
    ) -> AwaitableOrValue[Iterable[NodeType]]:
        """Resolve a list of nodes.

        This method *should* be defined by anyone implementing the `Node` interface.

        Args:
            info:
                The strawberry execution info resolve the type name from
            node_ids:
                Optional list of ids that, when provided, should be used to filter
                the results to only contain the nodes of those ids. When empty,
                all nodes of this type shall be returned.

        Returns:
            An iterable of resolved nodes.

        """
        raise NotImplementedError  # pragma:nocover

    @overload
    @classmethod
    @abc.abstractmethod
    def resolve_node(
        cls: Type[NodeType],
        node_id: str,
        *,
        info: Optional[Info] = ...,
        required: Literal[True],
    ) -> AwaitableOrValue[NodeType]:
        ...

    @overload
    @classmethod
    @abc.abstractmethod
    def resolve_node(
        cls: Type[NodeType],
        node_id: str,
        *,
        info: Optional[Info] = ...,
        required: bool = ...,
    ) -> AwaitableOrValue[Optional[NodeType]]:
        ...

    @classmethod
    @abc.abstractmethod
    def resolve_node(
        cls,
        node_id: str,
        *,
        info: Optional[Info] = None,
        required: bool = False,
    ):
        """Resolve a node given its id.

        This method *should* be defined by anyone implementing the `Node` interface.

        Args:
            info:
                The strawberry execution info resolve the type name from
            node_id:
                The id of the node to be retrieved
            required:
                if the node is required or not to exist. If not, then None should
                be returned if it doesn't exist. Otherwise an exception should be raised.

        Returns:
            The resolved node or None if it was not found

        """
        raise NotImplementedError  # pragma:nocover


@strawberry.type(description="Information to aid in pagination.")
class PageInfo:
    """Information to aid in pagination.

    Attributes:
        has_next_page:
            When paginating forwards, are there more items?
        has_previous_page:
            When paginating backwards, are there more items?
        start_cursor:
            When paginating backwards, the cursor to continue
        end_cursor:
            When paginating forwards, the cursor to continue

    """

    has_next_page: bool = strawberry.field(
        description="When paginating forwards, are there more items?",
    )
    has_previous_page: bool = strawberry.field(
        description="When paginating backwards, are there more items?",
    )
    start_cursor: Optional[str] = strawberry.field(
        description="When paginating backwards, the cursor to continue.",
    )
    end_cursor: Optional[str] = strawberry.field(
        description="When paginating forwards, the cursor to continue.",
    )


@strawberry.type(description="An edge in a connection.")
class Edge(Generic[NodeType]):
    """An edge in a connection.

    Attributes:
        cursor:
            A cursor for use in pagination
        node:
            The item at the end of the edge

    """

    cursor: str = strawberry.field(
        description="A cursor for use in pagination",
    )
    node: NodeType = strawberry.field(
        description="The item at the end of the edge",
    )

    @classmethod
    def from_node(cls, node: NodeType, *, cursor: Any = None):
        return cls(cursor=to_base64(connection_typename, cursor), node=node)


@strawberry.type(description="A connection to a list of items.")
class Connection(Generic[NodeType]):
    """A connection to a list of items.

    Attributes:
        page_info:
            Pagination data for this connection
        edges:
            Contains the nodes in this connection
        total_count:
            Total quantity of existing nodes

    """

    EDGE_CLASS: ClassVar[Type["Edge"]] = Edge

    page_info: PageInfo = strawberry.field(
        description="Pagination data for this connection",
    )
    edges: List[Edge[NodeType]] = strawberry.field(
        description="Contains the nodes in this connection",
    )
    total_count: Optional[int] = strawberry.field(
        description="Total quantity of existing nodes",
        default=None,
    )

    @classmethod
    def from_nodes(
        cls,
        nodes: Iterable[Any],
        *,
        total_count: Optional[int] = None,
        before: Optional[str] = None,
        after: Optional[str] = None,
        first: Optional[int] = None,
        last: Optional[int] = None,
    ):
        """Resolve a connection from the list of nodes.

        This uses the described Relay Pagination algorithm_

        Args:
            info:
                The strawberry execution info resolve the type name from
            nodes:
                An iterable of nodes to transform to a connection
            total_count:
                Optionally provide a total count so that the connection
                doesn't have to calculate it. Might be useful for some ORMs
                for performance reasons.
            before:
                Returns the items in the list that come before the specified cursor
            after:
                Returns the items in the list that come after the specified cursor
            first:
                Returns the first n items from the list
            last:
                Returns the items in the list that come after the specified cursor

        Returns:
            The resolved `Connection`

        .. _Relay Pagination algorithm:
            https://relay.dev/graphql/connections.htm#sec-Pagination-algorithm

        """
        if total_count is None:
            # Support ORMs that define .count() (e.g. django)
            try:
                total_count = int(nodes.count())  # type:ignore
            except (AttributeError, ValueError, TypeError):
                if isinstance(nodes, Sized):
                    total_count = len(nodes)

        max_results = config.RELAY_MAX_RESULTS
        if max_results is None:
            max_results = math.inf

        start = 0
        end = total_count if total_count is not None else math.inf

        if after:
            after_type, after_parsed = from_base64(after)
            assert after_type == connection_typename
            start = int(after_parsed) + 1
        if before:
            before_type, before_parsed = from_base64(before)
            assert before_type == connection_typename
            end = int(before_parsed)

        if end is None:
            end = max_results

        if isinstance(first, int):
            if first < 0:
                raise ValueError("Argument 'first' must be a non-negative integer.")

            if first > max_results:
                raise ValueError(f"Argument 'first' cannot be higher than {max_results}.")

            end = min(end, start + first)
        if isinstance(last, int):
            if last < 0:
                raise ValueError("Argument 'last' must be a non-negative integer.")

            if last > max_results:
                raise ValueError(f"Argument 'last' cannot be higher than {max_results}.")

            if end == math.inf:
                raise ValueError("Cannot use last with unlimited results")

            start = max(start, end - last)

        # If at this point end is still inf, consider it to be start + max_results
        if end == math.inf:
            end = start + max_results

        expected = end - start
        # If no parameters are given, end could be total_results at this point.
        # Make sure we don't exceed max_results in here
        if expected > max_results:
            end = start + max_results
            expected = end - start

        # Overfetch by 1 to check if we have a next result
        edges = [
            cls.EDGE_CLASS.from_node(v, cursor=start + i)
            for i, v in enumerate(cast(Sequence, nodes)[start : end + 1])  # noqa:E203
        ]

        # Remove the overfetched result
        if len(edges) == expected + 1:
            edges = edges[:-1]
            has_next_page = True
        else:
            has_next_page = False

        page_info = PageInfo(
            start_cursor=edges[0].cursor if edges else None,
            end_cursor=edges[-1].cursor if edges else None,
            has_previous_page=start > 0,
            has_next_page=has_next_page,
        )

        return cls(
            edges=edges,
            page_info=page_info,
            total_count=total_count,
        )


Node.CONNECTION_CLASS = Connection


class RelayField(StrawberryField):
    """Base relay field, containing utilities for both Node and Connection fields."""

    default_args: Dict[str, StrawberryArgument]

    def __init__(self, *args, **kwargs):
        default_args = getattr(self.__class__, "default_args", {})
        if isinstance(default_args, dict):
            self.default_args = default_args.copy()
        base_resolver = kwargs.pop("base_resolver", None)
        super().__init__(*args, **kwargs)
        if base_resolver:
            self.__call__(base_resolver)

    @property
    def arguments(self) -> List[StrawberryArgument]:
        args = {
            **self.default_args,
            **{arg.python_name: arg for arg in super().arguments},
        }
        return list(args.values())

    @functools.cached_property
    def is_basic_field(self):
        return False

    @functools.cached_property
    def is_optional(self):
        return isinstance(self.type, StrawberryOptional)

    @functools.cached_property
    def is_list(self):
        type_ = self.type
        if isinstance(type_, StrawberryOptional):
            type_ = type_.of_type

        return isinstance(type_, StrawberryList)


class NodeField(RelayField):
    """Relay Node field.

    Do not instantiate this directly. Instead, use `@relay.node`

    """

    @property
    def default_args(self) -> Dict[str, StrawberryArgument]:
        if self.base_resolver:
            return {}

        if self.is_list:
            return {
                "ids": StrawberryArgument(
                    python_name="ids",
                    graphql_name=None,
                    type_annotation=StrawberryAnnotation(List[GlobalID]),
                    description="The IDs of the objects.",
                ),
            }
        else:
            return {
                "id": StrawberryArgument(
                    python_name="id",
                    graphql_name=None,
                    type_annotation=StrawberryAnnotation(GlobalID),
                    description="The ID of the object.",
                ),
            }

    def __call__(self, resolver):
        raise TypeError("NodeField cannot have a resolver, use a common field instead.")

    def get_result(
        self,
        source: Any,
        info: Info,
        args: List[Any],
        kwargs: Dict[str, Any],
    ) -> AwaitableOrValue[Any]:
        if self.is_list:
            return self.resolve_nodes(source, info, args, kwargs)
        else:
            return self.resolve_node(source, info, args, kwargs)

    def resolve_node(
        self,
        source: Any,
        info: Info,
        args: List[Any],
        kwargs: Dict[str, Any],
    ) -> AwaitableOrValue[Optional[Node]]:
        gid = kwargs["id"]
        assert isinstance(gid, GlobalID)
        return gid.resolve_type(info).resolve_node(
            gid.node_id,
            info=info,
            required=not self.is_optional,
        )

    def resolve_nodes(
        self,
        source: Any,
        info: Info,
        args: List[Any],
        kwargs: Dict[str, Any],
    ) -> AwaitableOrValue[Iterable[Node]]:
        nodes_map: Dict[Type[Node], List[str]] = {}
        for gid in kwargs["ids"]:
            node_t = gid.resolve_type(info)
            nodes_map.setdefault(node_t, []).append(gid.node_id)

        if len(nodes_map) == 0:
            return []
        if len(nodes_map) > 1:
            # FIXME: Maybe we want to support this in the future?
            raise TypeError("More than one node type found...")

        node_t, ids = next(iter(nodes_map.items()))
        return node_t.resolve_nodes(info=info, node_ids=ids)


class ConnectionField(RelayField):
    """Relay Connection field.

    Do not instantiate this directly. Instead, use `@relay.connection`

    """

    default_args: Dict[str, StrawberryArgument] = {
        "before": StrawberryArgument(
            python_name="before",
            graphql_name=None,
            type_annotation=StrawberryAnnotation(Optional[str]),
            description="Returns the items in the list that come before the specified cursor.",
            default=None,
        ),
        "after": StrawberryArgument(
            python_name="after",
            graphql_name=None,
            type_annotation=StrawberryAnnotation(Optional[str]),
            description="Returns the items in the list that come after the specified cursor.",
            default=None,
        ),
        "first": StrawberryArgument(
            python_name="first",
            graphql_name=None,
            type_annotation=StrawberryAnnotation(Optional[int]),
            description="Returns the first n items from the list.",
            default=None,
        ),
        "last": StrawberryArgument(
            python_name="last",
            graphql_name=None,
            type_annotation=StrawberryAnnotation(Optional[int]),
            description="Returns the items in the list that come after the specified cursor.",
            default=None,
        ),
    }

    def __call__(self, resolver: Callable[..., Iterable[Node]]):
        namespace = sys.modules[resolver.__module__].__dict__
        nodes_type = resolver.__annotations__.get("return")
        if not nodes_type:
            raise TypeError("Connection nodes resolver needs a return type decoration.")

        if isinstance(nodes_type, str):
            nodes_type = ForwardRef(nodes_type, is_argument=False)
        resolved = _eval_type(nodes_type, namespace, None)
        origin = get_origin(resolved)
        if not origin or (not isinstance(origin, type) and not issubclass(origin, Iterable)):
            raise TypeError(
                "Connection nodes resolver needs a decoration that is a subclass of Iterable, "
                "like `Iterable[<NodeType>]`, `List[<NodeType>]`, etc"
            )

        ntype = get_args(resolved)[0]
        if isinstance(ntype, LazyType):
            ntype = ntype.resolve_type()

        type_override = StrawberryAnnotation(
            ntype.CONNECTION_CLASS[ntype],
            namespace=namespace,
        ).resolve()

        resolver = StrawberryResolver(resolver, type_override=type_override)
        return super().__call__(resolver)

    @property
    def type(self) -> Union[StrawberryType, type]:  # noqa:A003
        # Strawberry 0.139+ resolves the field annotation first, but we need to use the resolver's
        # type here because it gets modified by us in the __call__ method
        if (
            self.base_resolver is not None
            and self.base_resolver.type is not None
            and not isinstance(self.base_resolver.type, StrawberryTypeVar)
        ):
            return self.base_resolver.type

        return super().type

    @type.setter
    def type(self, type_: Any) -> None:  # noqa:A003
        super(ConnectionField, self.__class__).type.fset(self, type_)  # type:ignore

    @functools.cached_property
    def resolver_args(self) -> Set[str]:
        resolver = self.base_resolver
        if not resolver:
            return set()

        if isinstance(resolver, StrawberryResolver):
            resolver = resolver.wrapped_func

        return set(inspect.signature(cast(Callable, resolver)).parameters.keys())

    def get_result(
        self,
        source: Any,
        info: Info,
        args: List[Any],
        kwargs: Dict[str, Any],
    ) -> AwaitableOrValue[Any]:
        type_def = info.return_type._type_definition  # type:ignore
        assert isinstance(type_def, TypeDefinition)

        field_type = type_def.type_var_map[NodeType]
        if isinstance(field_type, LazyType):
            field_type = field_type.resolve_type()

        if self.base_resolver is not None:
            # If base_resolver is not self.conn_resolver, then it is defined to something
            assert self.base_resolver

            resolver_args = self.resolver_args
            resolver_kwargs = {
                # Consider both args not in default args and the ones specified by the resolver,
                # in case they want to check "first"/"last"/"before"/"after"
                k: v
                for k, v in kwargs.items()
                if k in resolver_args
            }
            nodes = self.base_resolver(*args, **resolver_kwargs)
        else:
            nodes = None

        nodes = self.resolve_nodes(source, info, args, kwargs, nodes=nodes)
        # This will be passed to the field cconnection resolver
        kwargs = {k: v for k, v in kwargs.items() if k in self.default_args}

        return cast(Node, field_type).resolve_connection(info=info, nodes=nodes, **kwargs)

    def resolve_nodes(
        self,
        source: Any,
        info: Info,
        args: List[Any],
        kwargs: Dict[str, Any],
        *,
        nodes: Optional[Iterable[Node]] = None,
    ) -> Optional[AwaitableOrValue[Iterable[Node]]]:
        return nodes


class InputMutationField(RelayField):
    """Relay Mutation field.

    Do not instantiate this directly. Instead, use `@relay.mutation`

    """

    default_args: Dict[str, StrawberryArgument] = {}

    def __call__(self, resolver: Callable[..., Iterable[Node]]):
        name = to_camel_case(resolver.__name__)
        cap_name = name[0].upper() + name[1:]
        namespace = sys.modules[resolver.__module__].__dict__
        annotations = resolver.__annotations__
        resolver = StrawberryResolver(resolver)

        args = resolver.arguments
        type_dict: dict[str, Any] = {
            "__doc__": f"Input data for `{name}` mutation",
            "__annotations__": {},
        }
        f_types = {}
        for arg in args:
            annotation = annotations[arg.python_name]
            if get_origin(annotation) is Annotated:
                directives = tuple(
                    d for d in get_args(annotation)[1:] if hasattr(d, "__strawberry_directive__")
                )
            else:
                directives = ()

            type_dict["__annotations__"][arg.python_name] = annotation
            arg_field = strawberry.field(
                name=arg.graphql_name,
                is_subscription=arg.is_subscription,
                description=arg.description,
                default=arg.default,
                directives=directives,
            )
            arg_field.graphql_name = arg.graphql_name
            f_types[arg_field] = arg.type_annotation
            type_dict[arg.python_name] = arg_field

        # TODO: We are not creating a type for the output payload, as it is not easy to
        # do that with the typing system. Is there a way to solve that automatically?
        new_type = strawberry.input(type(f"{cap_name}Input", (), type_dict))
        self.default_args["input"] = StrawberryArgument(
            python_name="input",
            graphql_name=None,
            type_annotation=StrawberryAnnotation(new_type, namespace=namespace),
            description=type_dict["__doc__"],
        )

        # FIXME: We need to set this after strawberry.input() or else it will have problems
        # with Annotated annotations for scalar types. Find out why in the future...
        for f, annotation in f_types.items():
            f.type = annotation

        return super().__call__(resolver)

    @property
    def arguments(self) -> List[StrawberryArgument]:
        return list(self.default_args.values())

    def get_result(
        self,
        source: Any,
        info: Info,
        args: List[Any],
        kwargs: Dict[str, Any],
    ) -> AwaitableOrValue[Any]:
        assert self.base_resolver
        input_obj = kwargs.pop("input")
        return self.base_resolver(*args, **kwargs, **vars(input_obj))


def node(
    *,
    name: Optional[str] = None,
    is_subscription: bool = False,
    description: Optional[str] = None,
    permission_classes: Optional[List[Type[BasePermission]]] = None,
    deprecation_reason: Optional[str] = None,
    default: Any = dataclasses.MISSING,
    default_factory: Union[Callable[..., object], object] = dataclasses.MISSING,
    metadata: Optional[Mapping[Any, Any]] = None,
    directives: Optional[Sequence[object]] = (),
    # This init parameter is used by pyright to determine whether this field
    # is added in the constructor or not. It is not used to change
    # any behavior at the moment.
    init: Literal[True, False, None] = None,
) -> Any:
    """Annotate a property to create a relay query field.

    Examples:
        Annotating something like this:

        >>> @strawberry.type
        >>> class X:
        ...     some_node: SomeType = relay.node(description="ABC")

        Will produce a query like this that returns `SomeType` given its id.

        ```
        query {
            someNode (id: ID) {
                id
                ...
            }
        }
        ```

    """
    return NodeField(
        python_name=None,
        graphql_name=name,
        type_annotation=None,
        description=description,
        is_subscription=is_subscription,
        permission_classes=permission_classes or [],
        deprecation_reason=deprecation_reason,
        default=default,
        default_factory=default_factory,
        metadata=metadata,
        directives=directives or (),
    )


@overload
def connection(
    *,
    resolver: Callable[[], _T],
    name: Optional[str] = None,
    is_subscription: bool = False,
    description: Optional[str] = None,
    init: Literal[False] = False,
    permission_classes: Optional[List[Type[BasePermission]]] = None,
    deprecation_reason: Optional[str] = None,
    default: Any = dataclasses.MISSING,
    default_factory: Union[Callable[..., object], object] = dataclasses.MISSING,
    metadata: Optional[Mapping[Any, Any]] = None,
    directives: Optional[Sequence[object]] = (),
) -> _T:
    ...


@overload
def connection(
    *,
    name: Optional[str] = None,
    is_subscription: bool = False,
    description: Optional[str] = None,
    init: Literal[True] = True,
    permission_classes: Optional[List[Type[BasePermission]]] = None,
    deprecation_reason: Optional[str] = None,
    default: Any = dataclasses.MISSING,
    default_factory: Union[Callable[..., object], object] = dataclasses.MISSING,
    metadata: Optional[Mapping[Any, Any]] = None,
    directives: Optional[Sequence[object]] = (),
) -> Any:
    ...


@overload
def connection(
    resolver: Union[StrawberryResolver, Callable, staticmethod, classmethod],
    *,
    name: Optional[str] = None,
    is_subscription: bool = False,
    description: Optional[str] = None,
    permission_classes: Optional[List[Type[BasePermission]]] = None,
    deprecation_reason: Optional[str] = None,
    default: Any = dataclasses.MISSING,
    default_factory: Union[Callable[..., object], object] = dataclasses.MISSING,
    metadata: Optional[Mapping[Any, Any]] = None,
    directives: Optional[Sequence[object]] = (),
) -> ConnectionField:
    ...


def connection(
    resolver=None,
    *,
    name: Optional[str] = None,
    is_subscription: bool = False,
    description: Optional[str] = None,
    permission_classes: Optional[List[Type[BasePermission]]] = None,
    deprecation_reason: Optional[str] = None,
    default: Any = dataclasses.MISSING,
    default_factory: Union[Callable[..., object], object] = dataclasses.MISSING,
    metadata: Optional[Mapping[Any, Any]] = None,
    directives: Optional[Sequence[object]] = (),
    # This init parameter is used by pyright to determine whether this field
    # is added in the constructor or not. It is not used to change
    # any behavior at the moment.
    init: Literal[True, False, None] = None,
) -> Any:
    """Annotate a property or a method to create a relay connection field.

    Relay connections_ are mostly used for pagination purposes. This decorator
    helps creating a complete relay endpoint that provides default arguments
    and has a default implementation for the connection slicing.

    Note that when setting a resolver to this field, it is expected for this
    resolver to return an iterable of the expected node type, not the connection
    itself. That iterable will then be paginated accordingly. So, the main use
    case for this is to provide a filtered iterable of nodes by using some custom
    filter arguments.

    Examples:
        Annotating something like this:

        >>> @strawberry.type
        >>> class X:
        ...     some_node: relay.Connection[SomeType] = relay.connection(description="ABC")
        ...
        ...     @relay.connection(description="ABC")
        ...     def get_some_nodes(self, age: int) -> Iterable[SomeType]:
        ...         ...

        Will produce a query like this:

        ```
        query {
            someNode (before: String, after: String, first: String, after: String, age: Int) {
                totalCount
                pageInfo {
                    hasNextPage
                    hasPreviousPage
                    startCursor
                    endCursor
                }
                edges {
                    cursor
                    node {
                        id
                        ...
                    }
                }
            }
        }
        ```

    .. _Relay connections:
        https://relay.dev/graphql/connections.htm

    """
    f = ConnectionField(
        python_name=None,
        graphql_name=name,
        type_annotation=None,
        description=description,
        is_subscription=is_subscription,
        permission_classes=permission_classes or [],
        deprecation_reason=deprecation_reason,
        default=default,
        default_factory=default_factory,
        metadata=metadata,
        directives=directives or (),
    )
    if resolver is not None:
        f = f(resolver)
    return f


@overload
def input_mutation(
    *,
    resolver: Callable[[], _T],
    name: Optional[str] = None,
    is_subscription: bool = False,
    description: Optional[str] = None,
    init: Literal[False] = False,
    permission_classes: Optional[List[Type[BasePermission]]] = None,
    deprecation_reason: Optional[str] = None,
    default: Any = dataclasses.MISSING,
    default_factory: Union[Callable[..., object], object] = dataclasses.MISSING,
    metadata: Optional[Mapping[Any, Any]] = None,
    directives: Optional[Sequence[object]] = (),
) -> _T:
    ...


@overload
def input_mutation(
    *,
    name: Optional[str] = None,
    is_subscription: bool = False,
    description: Optional[str] = None,
    init: Literal[True] = True,
    permission_classes: Optional[List[Type[BasePermission]]] = None,
    deprecation_reason: Optional[str] = None,
    default: Any = dataclasses.MISSING,
    default_factory: Union[Callable[..., object], object] = dataclasses.MISSING,
    metadata: Optional[Mapping[Any, Any]] = None,
    directives: Optional[Sequence[object]] = (),
) -> Any:
    ...


@overload
def input_mutation(
    resolver: Union[StrawberryResolver, Callable, staticmethod, classmethod],
    *,
    name: Optional[str] = None,
    is_subscription: bool = False,
    description: Optional[str] = None,
    permission_classes: Optional[List[Type[BasePermission]]] = None,
    deprecation_reason: Optional[str] = None,
    default: Any = dataclasses.MISSING,
    default_factory: Union[Callable[..., object], object] = dataclasses.MISSING,
    metadata: Optional[Mapping[Any, Any]] = None,
    directives: Optional[Sequence[object]] = (),
) -> InputMutationField:
    ...


def input_mutation(
    resolver=None,
    *,
    name: Optional[str] = None,
    is_subscription: bool = False,
    description: Optional[str] = None,
    permission_classes: Optional[List[Type[BasePermission]]] = None,
    deprecation_reason: Optional[str] = None,
    default: Any = dataclasses.MISSING,
    default_factory: Union[Callable[..., object], object] = dataclasses.MISSING,
    metadata: Optional[Mapping[Any, Any]] = None,
    directives: Optional[Sequence[object]] = (),
    # This init parameter is used by pyright to determine whether this field
    # is added in the constructor or not. It is not used to change
    # any behavior at the moment.
    init: Literal[True, False, None] = None,
) -> Any:
    """Annotate a property or a method to create an input mutation field.

    The difference from this mutation to the default one from strawberry is that
    all arguments found in the resolver will be converted to a single input type,
    named using the mutation name, capitalizing the first letter and append "Input"
    at the end. e.g. `doSomeMutation` will generate an input type `DoSomeMutationInput`.

    Examples:
        Annotating something like this:

        >>> @strawberry.type
        ... class CreateUserPayload:
        ...     user: UserType
        ...
        >>> @strawberry.mutation
        >>> class X:
        ...     @relay.input_mutation
        ...     def create_user(self, name: str, age: int) -> UserPayload:
        ...         ...

        Will create a type and an input type like

        ```
        input CreateUserInput {
            name: String!
            age: Int!
        }

        mutation {
            createUser (input: CreateUserInput!) {
                user: UserType
            }
        }
        ```

    """
    f = InputMutationField(
        python_name=None,
        graphql_name=name,
        type_annotation=None,
        description=description,
        is_subscription=is_subscription,
        permission_classes=permission_classes or [],
        deprecation_reason=deprecation_reason,
        default=default,
        default_factory=default_factory,
        metadata=metadata,
        directives=directives or (),
    )
    if resolver is not None:
        f = f(resolver)
    return f
